from sqlalchemy import Integer, String, Column, DateTime
from ..common import Base


class Summoner(Base):
    __tablename__ = 'summoner'

    account_id = Column(String, primary_key=True)

    summoner_id = Column(String)
    puuid = Column(String)
    summoner_name = Column(String)
    profile_icon_id = Column(Integer)
    revision_date = Column(DateTime)
    summoner_level = Column(Integer)


from sqlalchemy import Column, String, DateTime, Integer, Boolean, ForeignKey
from ..common import Base
from datetime import datetime


class League(Base):
    __tablename__ = 'league'

    league_id = Column(String, primary_key=True)
    tier = Column(String)
    name = Column(String)
    queue = Column(String)


class LeagueHistory(Base):
    __tablename__ = 'league_history'

    summoner_id = Column(String, primary_key=True)
    league_id = Column(String, primary_key=True)
    wins = Column(Integer, primary_key=True)
    losses = Column(Integer, primary_key=True)

    event_time = Column(DateTime, default=datetime.utcnow)
    league_points = Column(Integer)


from sqlalchemy import Column, String, DateTime, Integer, Boolean, ForeignKey
from ..common import Base


class Match(Base):
    __tablename__ = 'matches'

    game_id = Column(Integer, primary_key=True)

    platform_id = Column(String)
    game_creation = Column(DateTime)
    game_duration = Column(Integer)
    queue_id = Column(String)
    map_id = Column(String)
    season_id = Column(String)
    game_version = Column(String)
    game_mode = Column(String)
    game_type = Column(String)


class Team(Base):
    __tablename__ = 'teams'

    game_id = Column(Integer, ForeignKey('matches.game_id', ondelete='CASCADE'), primary_key=True)
    team_id = Column(String, primary_key=True)

    win = Column(String)
    first_blood = Column(Boolean)
    first_tower = Column(Boolean)
    first_inhibitor = Column(Boolean)
    first_baron = Column(Boolean)
    first_dragon = Column(Boolean)
    first_rift_herald = Column(Boolean)
    tower_kills = Column(Integer)
    inhibitor_kills = Column(Integer)
    baron_kills = Column(Integer)
    dragon_kills = Column(Integer)
    vilemaw_kills = Column(Integer)
    rift_herald_kills = Column(Integer)
    dominion_victory_score = Column(Integer)


class Ban(Base):
    __tablename__ = 'bans'

    game_id = Column(Integer, ForeignKey('matches.game_id', ondelete='CASCADE'), primary_key=True)
    team_id = Column(String, primary_key=True)
    pick_turn = Column(Integer, primary_key=True)
    champion_id = Column(Integer, primary_key=True)


class Participant(Base):
    __tablename__ = 'participants'

    game_id = Column(Integer, ForeignKey('matches.game_id', ondelete='CASCADE'), primary_key=True)
    participant_id = Column(String, primary_key=True)

    platform_id = Column(String)
    account_id = Column(String)
    summoner_name = Column(String)
    summoner_id = Column(String)
    current_platform_id = Column(String)
    current_account_id = Column(String)
    match_history_uri = Column(String)
    profile_icon = Column(Integer)


class Stats(Base):
    __tablename__ = 'stats'

    game_id = Column(Integer, ForeignKey('matches.game_id', ondelete='CASCADE'), primary_key=True)
    team_id = Column(String, primary_key=True)
    participant_id = Column(String, primary_key=True)
    
    champion_id = Column(Integer)
    spell1_id = Column(Integer)
    spell2_id = Column(Integer)
    win = Column(Boolean)
    item0 = Column(Integer)
    item1 = Column(Integer)
    item2 = Column(Integer)
    item3 = Column(Integer)
    item4 = Column(Integer)
    item5 = Column(Integer)
    item6 = Column(Integer)
    kills = Column(Integer)
    deaths = Column(Integer)
    assists = Column(Integer)
    largest_killing_spree = Column(Integer)
    largest_multi_kill = Column(Integer)
    killing_sprees = Column(Integer)
    longest_time_spent_living = Column(Integer)
    double_kills = Column(Integer)
    triple_kills = Column(Integer)
    quadra_kills = Column(Integer)
    penta_kills = Column(Integer)
    unreal_kills = Column(Integer)
    total_damage_dealt = Column(Integer)
    magic_damage_dealt = Column(Integer)
    physical_damage_dealt = Column(Integer)
    true_damage_dealt = Column(Integer)
    largest_critical_strike = Column(Integer)
    total_damage_dealt_to_champions = Column(Integer)
    magic_damage_dealt_to_champions = Column(Integer)
    physical_damage_dealt_to_champions = Column(Integer)
    true_damage_dealt_to_champions = Column(Integer)
    total_heal = Column(Integer)
    total_units_healed = Column(Integer)
    damage_self_mitigated = Column(Integer)
    damage_dealt_to_objectives = Column(Integer)
    damage_dealt_to_turrets = Column(Integer)
    vision_score = Column(Integer)
    time_c_cing_others = Column(Integer)
    total_damage_taken = Column(Integer)
    magical_damage_taken = Column(Integer)
    physical_damage_taken = Column(Integer)
    true_damage_taken = Column(Integer)
    gold_earned = Column(Integer)
    gold_spent = Column(Integer)
    turret_kills = Column(Integer)
    inhibitor_kills = Column(Integer)
    total_minions_killed = Column(Integer)
    neutral_minions_killed = Column(Integer)
    neutral_minions_killed_team_jungle = Column(Integer)
    neutral_minions_killed_enemy_jungle = Column(Integer)
    total_time_crowd_control_dealt = Column(Integer)
    champ_level = Column(Integer)
    vision_wards_bought_in_game = Column(Integer)
    sight_wards_bought_in_game = Column(Integer)
    wards_placed = Column(Integer)
    wards_killed = Column(Integer)
    first_blood_kill = Column(Boolean)
    first_blood_assist = Column(Boolean)
    first_tower_kill = Column(Boolean)
    first_tower_assist = Column(Boolean)
    first_inhibitor_kill = Column(Boolean)
    first_inhibitor_assist = Column(Boolean)
    combat_player_score = Column(Integer)
    objective_player_score = Column(Integer)
    total_player_score = Column(Integer)
    total_score_rank = Column(Integer)
    player_score0 = Column(Integer)
    player_score1 = Column(Integer)
    player_score2 = Column(Integer)
    player_score3 = Column(Integer)
    player_score4 = Column(Integer)
    player_score5 = Column(Integer)
    player_score6 = Column(Integer)
    player_score7 = Column(Integer)
    player_score8 = Column(Integer)
    player_score9 = Column(Integer)
    perk0 = Column(Integer)
    perk0_var1 = Column(Integer)
    perk0_var2 = Column(Integer)
    perk0_var3 = Column(Integer)
    perk1 = Column(Integer)
    perk1_var1 = Column(Integer)
    perk1_var2 = Column(Integer)
    perk1_var3 = Column(Integer)
    perk2 = Column(Integer)
    perk2_var1 = Column(Integer)
    perk2_var2 = Column(Integer)
    perk2_var3 = Column(Integer)
    perk3 = Column(Integer)
    perk3_var1 = Column(Integer)
    perk3_var2 = Column(Integer)
    perk3_var3 = Column(Integer)
    perk4 = Column(Integer)
    perk4_var1 = Column(Integer)
    perk4_var2 = Column(Integer)
    perk4_var3 = Column(Integer)
    perk5 = Column(Integer)
    perk5_var1 = Column(Integer)
    perk5_var2 = Column(Integer)
    perk5_var3 = Column(Integer)
    perk_primary_style = Column(Integer)
    perk_sub_style = Column(Integer)
    stat_perk0 = Column(Integer)
    stat_perk1 = Column(Integer)
    stat_perk2 = Column(Integer)
    role = Column(String)
    lane = Column(String)


from sqlalchemy import event, DDL

v_summoner_matches = DDL(
    '''
    create view if not exists v_summoner_matches as
    select
        a.game_id,
        a.summoner_name,
        d.name,
        e.description,
        b.season_id,
        b.game_creation,
        b.game_duration,
        c.win,
        c.kills,
        c.deaths,
        c.assists,
        c.role,
        c.lane
    from
        participants a
        join matches b on (a.game_id = b.game_id)
        join stats c on(a.game_id = c.game_id and a.participant_id = c.participant_id)
        join champions d on(c.champion_id = d.champion_id)
        join  queues e on(b.queue_id = e.queue_id)
    order by game_creation desc;
    '''
)

v_summoner_recently_played_champions = DDL(
    '''
    create view if not exists v_summoner_recently_played_champions  as
    select
        summoner_name,
        description,
        name,
        count(*) played
    from
    (
             select
                game_id,
                summoner_name,
                name,
                description,
                game_creation,
                rank() over (partition by summoner_name, description order by game_creation desc) as pos
             from v_summoner_matches
    )
    where
        pos <= 20
    group by summoner_name, description, name;
    '''
)



import requests
import pandas as pd
import re
import json
import logging  # https://docs.python.org/3/howto/logging.html#logging-basic-tutorial
from collections import deque
from typing import Dict, Union
from time import sleep
from datetime import datetime
from enum import Enum


logging.basicConfig(filename='maokai.log', level=logging.INFO, format='%(asctime)s %(message)s',
                    datefmt='%d.%m.%Y %H:%M:%S')

regions = {
    'EUW': 'https://euw1.api.riotgames.com'
}

endpoints = {
    'summoner_v4': {
        'by_account': '{base_uri}/lol/summoner/v4/summoners/by-account/{encryptedAccountId}',
        'by_name': '{base_uri}/lol/summoner/v4/summoners/by-name/{summonerName}',
        'by_puuid': '{base_uri}/lol/summoner/v4/summoners/by-puuid/{encryptedPUUID}',
        'by_id': '{base_uri}/lol/summoner/v4/summoners/{encryptedSummonerId}'
    },
    'match_v4': {
        'list': '{base_uri}/lol/match/v4/matchlists/by-account/{encryptedAccountId}',
        'details': '{base_uri}/lol/match/v4/matches/{matchId}',
        'timeline': '{base_uri}/lol/match/v4/timelines/by-match/{matchId}'
    },
    'league_v4': {
        'challenger': '{base_uri}/lol/league/v4/challengerleagues/by-queue/{queue_type}',
        'grandmaster': '{base_uri}/lol/league/v4/grandmasterleagues/by-queue/{queue_type}',
        'master': '{base_uri}/lol/league/v4/masterleagues/by-queue/{queue_type}',
        'entries': '{base_uri}/lol/league/v4/entries/by-summoner/{encryptedSummonerId}',
        'league': '{base_uri}/lol/league/v4/leagues/{leagueId}'
    },
    'ddragon': {
        'champion': 'http://ddragon.leagueoflegends.com/cdn/{local}/data/en_US/champion.json',
        'version': 'https://ddragon.leagueoflegends.com/api/versions.json',
        'queues': 'http://static.developer.riotgames.com/docs/lol/queues.json'
    }
}

status_codes = {
    400: 'bad request',
    401: 'unauthorized',
    403: 'forbidden',
    404: 'data not found',
    405: 'method not allowed',
    415: 'unsupported media type',
    429: 'rate limit exceeded',
    500: 'internal server error',
    504: 'service unavaiable'
}


class QueueType(Enum):
    RANKED_SOLO = 'RANKED_SOLO_5x5'
    RANKED_FLEX = 'RANKED_FLEX_SR'


class RiotApi:
    def __init__(self, api_key: str, region: str = 'EUW') -> None:
        if region not in regions:
            raise ValueError('{region} is not a valid region'.format(region=region))

        self._base_uri = regions[region]
        self._key = api_key
        self._header = {
            'Origin': 'https://developer.riotgames.com',
            'X-Riot-Token': self._key
        }
        self._query_delay_time = 0.1
        self._version = self.get_latest_version()
        self._api_calls = deque(maxlen=100)

    @staticmethod
    def __snake_case(camel_case: str) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', camel_case).lower()

    def __post_query(self, query: str) -> Dict:
        sleep(self._query_delay_time)
        r = requests.get(query, headers=self._header)

        # query was succesful 
        if r.status_code == 200:
            self._api_calls.append('{time}: {query}'.format(time=datetime.now(), query=query))
            return r.json()

        # something went wrong...
        elif r.status_code in (400, 401, 403, 404, 405, 415, 500, 503, 504):
            raise Exception('{0} {1}: {2}'.format(r.status_code, status_codes[r.status_code], query))

        # bad request
        elif r.status_code == 429:
            logging.warning('Rate limit exceeded. Sleep 121 seconds.')
            sleep(125)
            try:
                r = requests.get(query, headers=self._header)
                return r.json()
            except Exception:
                if r.status_code != 200:
                    raise Exception('Repeated Error after rate limited exceeded and 125 seconds timeout')

        # bad request, unknown status code
        else:
            raise Exception('unknown error code {0} on requesting {1}'.format(r.status_code, query))

    def get_summoner(self, **kwargs) -> pd.DataFrame:
        """Query the riot api for a specific summoner and returns Dataframe.

        One of the following parameters have to be provided:
        :param summoner_id: query summoner by summoner_id.
        :param account_id: query summoner by account_id.
        :param summoner_name: query summoner by summoner_name.
        :param puuid: query summoner by puuid.

        Raises:
            ValueError: thrown if none of [summoner_id, account_id, summoner_name, puuid] gets provided.

        Returns:
            pd.Dataframe: contains information about the summoner by the riot api.
        """

        if 'summoner_id' in kwargs:
            query = endpoints['summoner_v4']['by_id'].format(base_uri=self._base_uri,
                                                             encryptedSummonerId=kwargs['summoner_id'])
        elif 'account_id' in kwargs:
            query = endpoints['summoner_v4']['by_account'].format(base_uri=self._base_uri,
                                                                  encryptedAccountId=kwargs['account_id'])
        elif 'summoner_name' in kwargs:
            query = endpoints['summoner_v4']['by_name'].format(base_uri=self._base_uri,
                                                               summonerName=kwargs['summoner_name'])
        elif 'puuid' in kwargs:
            query = endpoints['summoner_v4']['by_puuid'].format(base_uri=self._base_uri,
                                                                encryptedPUUID=kwargs['puuid'])
        else:
            raise ValueError('[summoner_id, account_id, summoner_name, puuid] is needed to get summoner data')

        result = self.__post_query(query)
        df_result = pd.json_normalize(result)
        df_result.columns = map(self.__snake_case, df_result.columns)
        df_result.rename({'id': 'summoner_id', 'name': 'summoner_name'}, axis=1, inplace=True)
        df_result.set_index('account_id', inplace=True)
        df_result['revision_date'] = pd.to_datetime(df_result['revision_date'], unit='ms')
        return df_result

    def get_match_history(self, **kwargs) -> pd.DataFrame:
        """Query the riot api for match history of summonner

        One of the following parameters have to be provided:
        :param summoner_id: query summoner by summoner_id.
        :param account_id: query summoner by account_id.
        :param summoner_name: query summoner by summoner_name.
        :param puuid: query summoner by puuid.

        optional kwargs:
        champion: champion id
        queue: queue id
        season: season id
        endTime: time in ms
        beginTime: time in ms
        endIndex: amount of games which get returned
        beginIndex: start index of games returned

        Raises:
            ValueError: thrown if none of [summoner_id, account_id, summoner_name, puuid] gets provided.

        Returns:
            pd.Dataframe: match history of summoner
        """
        optional = ['champion', 'queue', 'season', 'endTime', 'beginTime', 'endIndex', 'beginIndex']
        if 'account_id' in kwargs:
            query = endpoints['match_v4']['list'].format(base_uri=self._base_uri,
                                                         encryptedAccountId=kwargs['account_id'])

            query_filter = ['{0}={1}'.format(k, v) for k, v in kwargs.items() if k in optional and v is not None]
            query = '{query}?{optional}'.format(query=query, optional='&'.join(query_filter))
            result = self.__post_query(query)['matches']

            df_matches = pd.json_normalize(result)
            df_matches.columns = map(self.__snake_case, df_matches.columns)
            return df_matches
        elif any(i in kwargs for i in ['summoner_name', 'summoner_id', 'puuid']):
            summoner = self.get_summoner(**kwargs).reset_index().iloc[0, :]
            return self.get_match_history(**{**summoner.to_dict(), **kwargs})
        else:
            raise ValueError(
                '[summoner_id, account_id, summoner_name, puuid] is needed to get match history of summoner')

    def get_match_details(self, match_id: Union[str, int]) -> Dict[str, pd.DataFrame]:
        query = endpoints['match_v4']['details'].format(base_uri=self._base_uri,
                                                        matchId=match_id)

        result = self.__post_query(query)
        frames = {}
        frames.update(self.__extract_match_data(result))
        frames.update(self.__extract_teams_data(result))
        frames.update(self.__extract_bans_data(result))
        frames.update(self.__extract_participants_data(result))
        frames.update(self.__extract_stats_data(result))
        return frames

    def get_timeline(self, match_id: Union[str, int]) -> Dict[str, pd.DataFrame]:
        query = endpoints['match_v4']['timeline'].format(base_uri=self._base_uri,
                                                         matchId=match_id)

        result = self.__post_query(query)

        df_participants = pd.DataFrame()
        df_events = pd.DataFrame()
        for frame in result['frames']:
            for participant in frame['participantFrames'].values():
                buffer = pd.json_normalize(participant, sep='_')
                buffer['timestamp'] = frame['timestamp']
                df_participants = df_participants.append(buffer, ignore_index=True)
            for event in frame['events']:
                buffer = pd.json_normalize(event, sep='_')
                df_events = df_events.append(buffer, ignore_index=True)

        df_participants.columns = map(self.__snake_case, df_participants.columns)
        df_participants['game_id'] = str(match_id)
        df_participants = df_participants.set_index(['game_id', 'timestamp', 'participant_id'])

        df_events.columns = map(self.__snake_case, df_events.columns)
        df_events.participant_id = df_events.participant_id.fillna('0')
        df_events['game_id'] = str(match_id)
        df_events['sequence'] = df_events.groupby(['game_id', 'timestamp', 'participant_id', 'type']).cumcount()
        df_events = df_events.set_index(['game_id', 'timestamp', 'participant_id', 'type'])

        frames = {}
        frames.update({'timeline_participants': df_participants})
        frames.update({'timeline_events': df_events})

        return frames

    def get_challenger_leaderboard(self, queue_type: QueueType) -> Dict[str, pd.DataFrame]:
        query = endpoints['league_v4']['challenger'].format(base_uri=self._base_uri,
                                                            queue_type=queue_type.value)
        result = self.__post_query(query)
        return self.__extract_leaderboard(result)

    def get_grandmaster_leaderboard(self, queue_type: QueueType) -> Dict[str, pd.DataFrame]:
        query = endpoints['league_v4']['grandmaster'].format(base_uri=self._base_uri,
                                                             queue_type=queue_type.value)
        result = self.__post_query(query)
        return self.__extract_leaderboard(result)

    def get_master_leaderboard(self, queue_type: QueueType) -> Dict[str, pd.DataFrame]:
        query = endpoints['league_v4']['master'].format(base_uri=self._base_uri,
                                                        queue_type=queue_type.value)
        result = self.__post_query(query)
        return self.__extract_leaderboard(result)

    def get_league(self, league_id: str) -> Dict[str, pd.DataFrame]:
        query = endpoints['league_v4']['league'].format(base_uri=self._base_uri,
                                                        leagueId=league_id)
        result = self.__post_query(query)

        league = pd.json_normalize(result).drop(columns='entries')
        league.columns = map(self.__snake_case, league.columns)
        league = league.set_index('league_id')

        entries = pd.json_normalize(result, record_path='entries', meta='leagueId')
        entries = entries.loc[:, ~entries.columns.str.startswith('miniSeries')]
        entries.columns = map(self.__snake_case, entries.columns)
        entries = entries.set_index('league_id')

        return {'league': league, 'entries': entries}

    def get_league_entries_of_summoner(self, **kwargs) -> pd.DataFrame:
        """Query the riot api for the league entries of a summoner.

        One of the following parameters have to be provided:
        :param summoner_id: query summoner by summoner_id.
        :param account_id: query summoner by account_id.
        :param summoner_name: query summoner by summoner_name.
        :param puuid: query summoner by puuid.

        Raises:
            ValueError: thrown if none of [summoner_id, account_id, summoner_name, puuid] gets provided.

        Returns:
            pd.Dataframe: contains information about league entries of the summoner by the riot api.
        """

        if 'summoner_id' in kwargs:
            query = endpoints['league_v4']['entries'].format(base_uri=self._base_uri,
                                                             encryptedSummonerId=kwargs['summoner_id'])
            result = self.__post_query(query)
            df_entries = pd.DataFrame()
            for entry in result:
                buffer = pd.json_normalize(entry, sep='_')
                df_entries = df_entries.append(buffer, ignore_index=True)

            df_entries.columns = map(self.__snake_case, df_entries.columns)
            df_entries = df_entries.set_index(['summoner_id', 'queue_type'])
            return df_entries

        elif any(i in kwargs for i in ['summoner_name', 'account_id', 'puuid']):
            summoner = self.get_summoner(**kwargs).reset_index().iloc[0, :]
            return self.get_league_entries_of_summoner(**{**summoner.to_dict(), **kwargs})
        else:
            raise ValueError(
                '[summoner_id, account_id, summoner_name, puuid] is needed to get league entries of summoner')

    def __extract_match_data(self, data: json) -> Dict[str, pd.DataFrame]:
        game = pd.json_normalize(data)
        game['gameCreation'] = pd.to_datetime(game['gameCreation'], unit='ms')
        game = game.drop(columns=['teams', 'participants', 'participantIdentities'])
        game.columns = map(self.__snake_case, game.columns)
        game = game.set_index('game_id')
        return {'matches': game}

    def __extract_teams_data(self, data: json) -> Dict[str, pd.DataFrame]:
        teams = pd.json_normalize(data, record_path=['teams'], meta='gameId', sep='_', max_level=0)
        teams = teams.drop(columns=['bans'])
        teams.columns = map(self.__snake_case, teams.columns)
        teams = teams.set_index(['game_id', 'team_id'])
        return {'teams': teams}

    def __extract_bans_data(self, data: json) -> Dict[str, pd.DataFrame]:
        try:
            bans = pd.json_normalize(data, record_path=['teams', 'bans'], meta=['gameId', ['teams', 'teamId']])
            bans = bans.rename(columns={'teams.teamId': 'teamId'})
            bans.columns = map(self.__snake_case, bans.columns)
            bans = bans.set_index(['game_id', 'team_id', 'pick_turn'])
            return {'bans': bans}
        except KeyError:
            return {'bans': pd.DataFrame()}

    def __extract_participants_data(self, data: json) -> Dict[str, pd.DataFrame]:
        participants = pd.json_normalize(data, record_path=['participantIdentities'], meta=['gameId'])
        participants.columns = map(lambda x: re.sub('player.', '', x), participants.columns)
        participants.columns = map(self.__snake_case, participants.columns)
        participants = participants.set_index(['game_id', 'participant_id'])
        return {'participants': participants}

    def __extract_stats_data(self, data: json) -> Dict[str, pd.DataFrame]:
        stats = pd.json_normalize(data, record_path=['participants'], meta=['gameId'])
        stats = stats.rename(columns={'timeline.lane': 'lane', 'timeline.role': 'role'})
        stats = stats.drop(columns=list(stats.filter(regex='timeline')))
        stats = stats.drop(columns=['participantId'])
        stats.columns = map(lambda x: re.sub('stats.', '', x), stats.columns)
        stats.columns = map(self.__snake_case, stats.columns)
        stats = stats.set_index(['game_id', 'team_id', 'participant_id'])
        return {'stats': stats}

    def __extract_leaderboard(self, data: json) -> Dict[str, pd.DataFrame]:
        df_entries = pd.json_normalize(data, record_path='entries', meta=['leagueId'])
        df_entries.columns = map(self.__snake_case, df_entries.columns)
        df_entries.set_index('league_id', 'summoner_id', inplace=True)

        df_entry = pd.DataFrame(data)
        df_entry.columns = map(self.__snake_case, df_entry.columns)
        df_entry = df_entry.drop(columns='entries').drop_duplicates()
        df_entry = df_entry.set_index('league_id')

        return {'league_entry': df_entry, 'league_entries': df_entries}

    def get_queue_types(self) -> pd.DataFrame:
        df_result = pd.read_json(endpoints['ddragon']['queues'])
        df_result.columns = map(self.__snake_case, df_result.columns)
        df_result = df_result.set_index('queue_id')
        return df_result

    def get_champion_data(self) -> pd.DataFrame:
        content = requests.get(endpoints['ddragon']['champion'].format(local=self._version)).json()
        table = pd.DataFrame()
        for value in content['data'].values():
            table = table.append(pd.json_normalize(value, sep='_'), ignore_index=True)
        table.columns = map(self.__snake_case, table)
        table.columns = [column.replace('stats_', '') for column in table.columns]
        table.rename(columns={'key': 'champion_id'}, inplace=True)
        table.drop(columns=['id'], inplace=True)
        table.set_index('champion_id', inplace=True)
        table['tags'] = table.tags.apply(lambda x: ', '.join(x))
        return table

    @staticmethod
    def get_versions() -> pd.DataFrame:
        versions = requests.get(endpoints['ddragon']['version']).json()
        return pd.DataFrame(versions)

    def get_latest_version(self) -> str:
        return self.get_versions().iloc[0, 0]



from ..api.riot import RiotApi, QueueType, logging

import pandas as pd
from datetime import datetime

from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import IntegrityError

from .models.views import v_summoner_matches, v_summoner_recently_played_champions
from .models.summoner import Summoner
from .models.match import Match
from .models.league import League, LeagueHistory
from .common import Base


class LeagueDB:
    def __init__(self, con: str, api_key: str):
        self.engine = create_engine(con)
        self.api = RiotApi(api_key)
        self.create_db_layout()

    def create_db_layout(self) -> None:
        Base.metadata.create_all(self.engine)
        self.create_static_data()
        self._create_views()

    def update_summoner(self, summoner_name: str, number_of_games: int = 100, champion_id: int = None,
                        season_id: str = None, patch: str = None, begin_time: datetime = None,
                        queue_id: int = None, get_timeline_data: bool = False) -> None:
        logging.info('update summoner: {0}'.format(summoner_name))
        with Session(self.engine) as session:
            try:
                # step 1: gather summoner data
                df_summoner = self.api.get_summoner(summoner_name=summoner_name)
                if df_summoner.empty:
                    logging.info('summoner with name {0} not found'.format(summoner_name))
                    return

                summoner = Summoner(**df_summoner.reset_index().iloc[0])
                session.merge(summoner)
                session.commit()

            except Exception as e:
                logging.error('error while gathering summoner data for summoner {0}'.format(summoner_name))
                logging.error(str(e))
                return

            try:
                # step 2: gather league information for summoner
                df_league_entries = self.api.get_league_entries_of_summoner(summoner_id=summoner.summoner_id)

                # check if league is already created
                for idx, entry in df_league_entries.reset_index().iterrows():
                    df_league = self.api.get_league(entry.league_id)['league']
                    session.merge(League(**df_league.reset_index().iloc[0]))
                    session.commit()

                    cols = ['summoner_id', 'league_id', 'league_points', 'wins', 'losses']
                    session.merge(LeagueHistory(**entry[cols]))
                    session.commit()

            except Exception as e:
                logging.error('error while gathering league data for summoner {0}'.format(summoner_name))
                logging.error(str(e))
                return

            try:
                # step 3: get match history of summoner
                matches = self.api.get_match_history(account_id=summoner.account_id, champion=champion_id,
                                                     endIndex=number_of_games, beginTime=begin_time, queue=queue_id)
                if matches.empty:
                    logging.info('no new matches for summoner {0}'.format(summoner_name))
                    return

                # check which matches are already stored in database
                query = select(Match).where(Match.game_id.in_(matches.game_id))
                db_matches = pd.read_sql(sql=query, con=session.bind)
                new_matches = matches[~matches.game_id.isin(db_matches.game_id)]

                # step 4 get match details of summoner
                logging.info('{0} out of {1} are new matches'.format(len(new_matches), len(matches)))
                for match in new_matches.game_id:
                    try:
                        details = self.api.get_match_details(match)
                        for name, table in details.items():
                            try:
                                table.to_sql(name=name, con=session.bind, if_exists='append')
                            except IntegrityError:
                                pass

                        if get_timeline_data:
                            timeline = self.api.get_timeline(match)
                            for name, table in timeline.items():
                                table = table.applymap(str)
                                table.to_sql(name=name, con=session.bind, if_exists='append')

                        logging.info('Merged {0} successfully'.format(match))
                    except Exception as e:
                        logging.error('error while gathering match details for game_id {0}'.format(match))
                        logging.error(str(e))
            except NoResultFound as e:
                logging.info('no new matches for summoner {0}'.format(summoner_name))
                pass
            except Exception as e:
                logging.error('error while gathering game_id data for summoner {0}'.format(summoner_name))
                logging.error(str(e))

    def _create_views(self) -> None:
        self.engine.execute(v_summoner_matches)
        self.engine.execute(v_summoner_recently_played_champions)

    def create_static_data(self) -> None:
        if not self.engine.has_table('leaderboard_flex') or not self.engine.has_table('leaderboard_solo'):
            logging.info('table "leaderboard_solo" and "leaderboard_flex" created')
            self._update_challenger_leaderboard()
        if not self.engine.has_table('champions'):
            logging.info('table "champions" created')
            self._update_champions()
        if not self.engine.has_table('queues'):
            logging.info('table "queues" created')
            self._update_queue_types()

    def _update_challenger_leaderboard(self) -> None:
        solo = self.api.get_challenger_leaderboard(QueueType.RANKED_SOLO)['league_entries']
        solo.to_sql(name='leaderboard_solo', con=self.engine, if_exists='replace')

        flex = self.api.get_challenger_leaderboard(QueueType.RANKED_FLEX)['league_entries']
        flex.to_sql(name='leaderboard_flex', con=self.engine, if_exists='replace')

    def _update_champions(self) -> None:
        champions = self.api.get_champion_data()
        champions.to_sql(name='champions', con=self.engine, if_exists='replace')

    def _update_queue_types(self) -> None:
        queues = self.api.get_queue_types()
        queues.to_sql(name='queues', con=self.engine, if_exists='replace')



from sqlalchemy import Column, Integer, String, ForeignKey
from ..common import Base


class TimelineParticipant(Base):
    __tablename__ = 'timeline_participants'

    game_id = Column(Integer, ForeignKey('matches.game_id', ondelete='CASCADE'), primary_key=True)
    timestamp = Column(Integer, primary_key=True)
    participant_id = Column(String, primary_key=True)

    current_gold = Column(Integer)
    total_gold = Column(Integer)
    level = Column(Integer)
    xp = Column(Integer)
    minions_killed = Column(Integer)
    jungle_minions_killed = Column(Integer)
    dominion_score = Column(Integer)
    team_score = Column(Integer)
    position_x = Column(Integer)
    position_y = Column(Integer)


class TimelineEvents(Base):
    __tablename__ = 'timeline_events'

    game_id = Column(Integer, ForeignKey('matches.game_id', ondelete='CASCADE'), primary_key=True)
    timestamp = Column(Integer, primary_key=True)
    participant_id = Column(String, primary_key=True)
    type = Column(String, primary_key=True)
    sequence = Column(Integer, primary_key=True)

    skill_slot = Column(String)
    level_up_type = Column(String)
    item_id = Column(String)
    ward_type = Column(String)
    creator_id = Column(String)
    killer_id = Column(String)
    victim_id = Column(String)
    assisting_participant_ids = Column(String)
    position_x = Column(String)
    position_y = Column(String)
    monster_type = Column(String)
    after_id = Column(String)
    before_id = Column(String)
    team_id = Column(String)
    building_type = Column(String)
    lane_type = Column(String)
    tower_type = Column(String)
    monster_sub_type = Column(String)